from abc import ABC, abstractmethod

class DataSourceInterface(ABC):
    @abstractmethod
    def get_datasource_name(self):
        """数据源名称"""
        pass

    @abstractmethod
    def get_port_tags(self):
        """获取产品标签"""
        pass

    @abstractmethod
    def get_port_inds(self):
        """获取产品指标"""
        pass

    @abstractmethod
    def get_asset_tags(self):
        """获取资产标签"""
        pass

    @abstractmethod
    def get_asset_inds(self):
        """获取产品指标"""
        pass

    @abstractmethod
    def get_basic_inds(self):
        """获取基础列指标"""
        pass

    @abstractmethod
    def get_data(self, run_method: str, data_subject: str, sec_ids: list=[], port_ids: list=[],
             tags: list=[], inds: list=[], start_date = None, end_date = None, penetrate=None):
        """获取资产标签、产品标签及指标信息"""
        pass

    @abstractmethod
    def get_repository_inds(self, run_method: str, repository: str, ind_list: list, params: dict):
        """获取指标库及指标信息"""
        pass


from muse.data_source.base import DataSourceInterface
from muse.data_source.ds_utils import filter_ports, filter_ports_sec, convert_decimal_to_float, get_all_unique_fields
from muse.muse_config import APP_LOCATE
import polars as pl

class DemoDataSource(DataSourceInterface):
    data_locate = APP_LOCATE + '/data/'

    def get_datasource_name(self):
        return 'Demo用数据源'

    def get_basic_inds(self):
        return ['资产代码', '资产名称', '资产标签', '产品代码', '产品简称', '产品标签', '主体代码', '主体名称', '主体标签', '日期']

    def get_port_tags(self):
        tags = pl.read_excel(self.data_locate + '产品标签列.xlsx')
        return tags.columns

    def get_port_inds(self):
        inds = pl.read_excel(self.data_locate + '产品指标列.xlsx')
        return inds.columns

    def get_asset_tags(self):
        tags = pl.read_excel(self.data_locate + '资产标签列.xlsx')
        return tags.columns

    def get_asset_inds(self):
        inds = pl.read_excel(self.data_locate + '资产指标列.xlsx')
        return inds.columns

    def get_data(self, run_method: str, data_subject: str, sec_ids: list = [], port_ids: list = [], tags: list = [],
                 inds: list = [], start_date=None, end_date=None, penetrate=None):
        if penetrate:
            penetrate = '全穿透'
        else:
            penetrate = '不穿透'
        pdf = None

        if data_subject == '持仓指标':
            asset_df = pl.read_excel(self.data_locate + '资产指标汇总.xlsx')
            asset_tags = pl.read_excel(self.data_locate + '资产标签.xlsx')
            asset_df = asset_df.join(asset_tags, on=['资产代码'], how='left')
            pdf = filter_ports_sec(pdf=asset_df, port_ids=port_ids, sec_ids=sec_ids, start_date=start_date,
                                   end_date=end_date)
            # 获得产品代码
            port_ids = pdf.select("产品代码").unique().to_series().to_list()
            ports = pl.read_excel(self.data_locate + '产品指标汇总.xlsx')
            port_tags = pl.read_excel(self.data_locate + '产品标签.xlsx')
            ports = ports.join(port_tags, on=['产品代码'], how='left')
            ports = filter_ports(pdf=ports, port_ids=port_ids, start_date=start_date, end_date=end_date)
            # 合并资产和产品
            pdf = pdf.join(ports, on=['产品代码', '日期'], how='left')
            pdf = convert_decimal_to_float(pdf)
        elif data_subject == '产品指标':
            ports = pl.read_excel(self.data_locate + '产品指标汇总.xlsx')
            port_tags = pl.read_excel(self.data_locate + '产品标签.xlsx')
            ports = ports.join(port_tags, on=['产品代码'], how='left')
            pdf = filter_ports_sec(pdf=ports, port_ids=port_ids, start_date=start_date, end_date=end_date)

        merged_cols = tags + inds
        if len(merged_cols) > 0:
            all_cols = get_all_unique_fields(merged_cols, self.get_basic_inds())
            existing_cols = [c for c in all_cols if c in pdf.columns]
            pdf = pdf.select(existing_cols)
        return pdf

    def get_repository_inds(self, run_method: str, repository: str, ind_list: list, params: dict):
        # 处理指标列表
        port_tags = params.get('port_tags', None)
        asset_tags = params.get('asset_tags', None)
        start_date = params.get('start_date', None)
        end_date = params.get('end_date', None)
        port_ids = params.get('port_ids', None)
        if port_ids:
            port_ids = port_ids.split(',').strip()
        else:
            port_ids = []
        sec_ids = params.get('sec_ids', None)
        if sec_ids:
            sec_ids = sec_ids.split(',').strip()
        else:
            sec_ids = []
        # 读取指标库
        rep_locate = f'{self.data_locate}指标库/{repository}.xlsx'
        # 选择合适的指标
        df = pl.read_excel(rep_locate)
        if len(ind_list) > 0:
            existing_cols = [col for col in set(self.get_basic_inds() + ind_list) if col in df.columns]
            df = df.select(existing_cols)
        # 过滤产品代码、资产代码和日期区间
        df = filter_ports_sec(df, port_ids, sec_ids, start_date, end_date)
        # 处理产品标签
        if port_tags:
            # 检查标签是否存在
            valid_port_tags = [tag for tag in port_tags if tag in self.get_port_tags()]
            if not valid_port_tags:
                raise ValueError(f"产品标签 {port_tags} 不在数据列中！")
            p_tags = pl.read_excel(f'{self.data_locate}产品标签.xlsx')
            p_tags = p_tags.select(['产品代码'] + valid_port_tags)
            df = df.join(p_tags, on=['产品代码'], how='left')
            # 动态构建 OR 条件
            port_condition = pl.any_horizontal(
                *[pl.col(tag) == "是" for tag in valid_port_tags]
            )
            df = df.filter(port_condition)
        # 处理资产标签
        if asset_tags:
            # 检查标签是否存在
            valid_asset_tags = [tag for tag in asset_tags if tag in self.get_asset_tags()]
            if not valid_asset_tags:
                raise ValueError(f"资产标签 {asset_tags} 不在数据列中！")
            # 动态构建 OR 条件
            a_tags = pl.read_excel(f'{self.data_locate}资产标签.xlsx')
            a_tags = a_tags.select(['资产代码'] + valid_asset_tags)
            df = df.join(a_tags, on=['资产代码'], how='left')
            asset_condition = pl.any_horizontal(
                *[pl.col(tag) == "是" for tag in valid_asset_tags]
            )
            df = df.filter(asset_condition)
        return df


# 指标提取引擎
# 根据金融对象.指标的方法提取金融对象的数值类指标
# 同一类金融对象需要一次性提取避免性能的额外开销
import logging
import polars as pl
import muse.data_interface as di

def indicator_handle(base, prop):
    if isinstance(base, pl.DataFrame) and prop not in base.columns:
        logging.error(f'调用点运算失败：{prop} 不在 {base}中')
        return prop, pl.DataFrame()
    return prop, base

def fetch_inds(params, port_ids=None, sec_ids=None, start_date=None, end_date=None, search_mode=None):
    # 第一个必填参数为指标库
    if len(params) < 3:
        logging.error(f'指标库调用函数需要3个参数，提供参数数量不正确')
        return pl.DataFrame()
    repository_name = params[0]
    ind_list = params[1]
    search_params = params[2]

    if port_ids:
        search_params['port_ids'] = port_ids
    if sec_ids:
        search_params['sec_ids'] = sec_ids
    if start_date:
        search_params['start_date'] = start_date
    if end_date:
        search_params['end_date'] = end_date

    return di.get_repository_inds(search_mode, repository_name, ind_list, search_params)


from setuptools import setup

setup(
    name="muse_lang",  # 包名称，pip install时使用的名字
    package_data={
        "muse": ["data/*", "data/**/*"],  # 包含data目录下的文件
    },
    version="0.4.2",      # 版本号
    description="Mini language for data analysis of asset management.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Brooks Lu",
    author_email="lxc@jixunet.com",
    url="https://e.coding.net/brookslu/muse/muse.git",
    install_requires=[   # 依赖的其他包
        "numpy>=1.24.4",
        "pandas>=2.0.3",
        "polars>=1.8.0",
        "fastexcel>=0.12.0"
    ],
    python_requires=">=3.8",  # Python版本要求
    classifiers=[       # 分类信息，可选
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


from muse.ast_struct import FunctionCall, Binary, AstVisitor, ArrayLiteral
from muse.engines import rule_engine
import pandas as pd
import polars as pl
import muse.data_interface as di
from muse.scanner import Op
import logging
import ast
import operator
import logging as logger
interpreter = AstVisitor()


def asset_tag(tags: list, props=None, sec_ids=None, start_date='', end_date='', drop_duplicate=False, search_mode='测试'):
    logging.debug('调用资产标签成功 {} with props {}'.format(tags, props))
    sec_list = []
    if sec_ids is not None:
        sec_list = [i.strip() for i in sec_ids.split(',')]
    ast = di.get_assets(search_mode, start_date, end_date, tags, sec_list, props)
    return ast


def issuer_tag(tags: list, props=None, start_date='', end_date='', drop_duplicate=False, search_mode='测试'):
    logging.debug('调用主体标签 {} with props {}'.format(tags, props))
    issuer_list = list()
    issuers = di.get_issuers(search_mode, start_date, end_date, tags, issuer_list, props)
    return issuers


def port_tag(tags: list, props=None, port_ids=None, start_date='', end_date ='', drop_duplicate=False, search_mode='测试'):
    logging.debug('调用产品标签  {} with props {}'.format(tags, props))
    # 对着所有的有效产品来处理
    port_list = list()
    if port_ids is not None:
        port_list = [i.strip() for i in port_ids.split(',')]
    port = di.get_ports(search_mode, start_date, end_date, tags, port_list, props)
    return port

def parse_tags_and_fields(params):
    tags = list()
    fields = list()
    pen = False
    drop_duplicates = True
    if len(params) > 0:
        if not isinstance(params[0], ArrayLiteral):
            for p in params:
                tags.append(interpreter.visit(p, None))
        else:
            tp = interpreter.visit(params[0], None)
            for tag in tp:
                tags.append(tag)
            if len(params) > 1:
                if isinstance(params[1], ArrayLiteral):
                    fp = interpreter.visit(params[1], None)
                    for field in fp:
                        fields.append(field)
                else:
                    result = interpreter.visit(params[1], None)
                    if result == '穿透' or result == '全穿透':
                        pen = '全穿透'
                    if result == '不去重':
                        drop_duplicates = False
            if len(params) > 2:
                result = interpreter.visit(params[2], None)
                if result == '穿透' or result == '全穿透':
                    pen = '全穿透'
                if result == '不去重':
                    drop_duplicates = False
            if len(params) > 3:
                result = interpreter.visit(params[3], None)
                if result == '穿透' or result == '全穿透':
                    pen = '全穿透'
                if result == '不去重':
                    drop_duplicates = False

    return tags, fields, pen, drop_duplicates

def in_tag(func1: FunctionCall, func2: FunctionCall, props=None, sec_ids=None, port_ids=None, start_date='', end_date='', search_mode='测试'):
    logging.debug('调用in标签成功 with func1: {} func2: {} props {}'.format(func1.name, func2.name, props))
    penetrate = '不穿透'
    # 解析第一个函数调用
    name1 = func1.name
    tags1, fields1, pen1, drop1 = parse_tags_and_fields(func1.params)

    # 解析第二个函数调用
    name2 = func2.name
    tags2, fields2, pen2, drop2 = parse_tags_and_fields(func2.params)


    # ps1里包含要取的字段
    if name2 == '产品' and name1 == '资产':
        asset_tags = tags1
        asset_fields = fields1
        asset_drop_duplicate = drop1
        penetrate = pen1

        port_tags = tags2
        port_fields = fields2
        port_drop_duplicate = drop2
    else:
        asset_tags = tags2
        asset_fields = fields2
        asset_drop_duplicate = drop2
        penetrate = pen2

        port_tags = tags1
        port_fields = fields1
        port_drop_duplicate = drop1

    if props is None:
        props = list()

    hld_props = list(set(asset_fields + port_fields + list(props)))
    hlds = di.get_hlds(search_mode, start_date, end_date, asset_tags, port_tags, list(hld_props), sec_ids=sec_ids, port_ids=port_ids, penetrate=penetrate)
    return hlds


def in_tag2(b1: Binary, func3: FunctionCall, props=None, sec_ids=None, port_ids=None, start_date='', end_date='', search_mode='测试'):
    logging.debug('调用2连in标签成功 with func1: {} func2: {} func3: {} props {}'.format(b1.exp1, b1.exp2, func3.name, props))
    penetrate = '不穿透'
    # 只有三种标签类型都在的情况下才能用这个函数来实现
    hlds = pd.DataFrame()
    if props is None:
        props = list()
    if b1.op == Op.In and isinstance(b1.exp1, FunctionCall) and isinstance(b1.exp2, FunctionCall):
        func1 = b1.exp1
        func2 = b1.exp2
        funcs = dict()
        hld_props =list()

        for f in [func1, func2, func3]:
            tags, a_props, pen, drop_duplicate = parse_tags_and_fields(f.params)
            a_props = list(set(a_props + list(props)))
            funcs[f.name] = [tags, a_props, pen, drop_duplicate]
            hld_props = list(set(hld_props + a_props))

        penetrate = funcs['资产'][2]
        asts = asset_tag(funcs['资产'][0], funcs['资产'][1], sec_ids, start_date, end_date, drop_duplicate=funcs['资产'][3])
        issuers = issuer_tag(funcs['主体'][0], funcs['主体'][1], start_date, end_date, drop_duplicate=funcs['主体'][3])
        ports = port_tag(funcs['产品'][0], funcs['产品'][1], port_ids, start_date, end_date, drop_duplicate=funcs['主体'][3])
        if isinstance(asts, pl.DataFrame) and asts.is_empty():
            return pl.DataFrame()
        if isinstance(ports, pl.DataFrame) and ports.is_empty():
            return pl.DataFrame()
        hlds = di.get_hlds2(search_mode, start_date, end_date, asts, issuers, ports, list(hld_props), penetrate=penetrate)
    return hlds


def in_tag3(df1: pl.DataFrame, df2: pl.DataFrame, props=None, sec_ids=None, port_ids=None, start_date='', end_date='', search_mode='测试'):
    asts = pl.DataFrame()
    ports = pl.DataFrame()
    if df1 is None:
        df1 = pl.DataFrame()
    if df2 is None:
        df2 = pl.DataFrame()

    if sec_ids is not None:
        sec_ids = [i.strip() for i in sec_ids.split(',')]
        asts = asset_tag([], props, sec_ids, start_date, end_date)
    else:
        if '资产代码' in df1:
            asts = df1
        if '资产代码' in df2:
            asts = df2

    if port_ids is not None:
        port_ids = [i.strip() for i in port_ids.split(',')]
        ports = port_tag([], props, port_ids, start_date, end_date)
    else:
        if '产品代码' in df1:
            ports = df1
        if '产品代码' in df2:
            ports = df2
    if isinstance(asts, pl.DataFrame) and asts.is_empty():
        return pl.DataFrame()
    if isinstance(ports, pl.DataFrame) and ports.is_empty():
        return pl.DataFrame()
    hlds = di.get_hlds3(search_mode, start_date, end_date, asts, ports, list(props))
    return hlds


def safe_eval_filter_rule(filter_rule: str):
    """安全地将类似Python的过滤表达式转换为Polars表达式"""
    # 支持的运算符映射
    op_map = {
        ast.Lt: operator.lt,
        ast.LtE: operator.le,
        ast.Gt: operator.gt,
        ast.GtE: operator.ge,
        ast.Eq: operator.eq,
        ast.NotEq: operator.ne,
        ast.Or: operator.or_,
        ast.And: operator.and_,
    }

    def _convert(node):
        if isinstance(node, ast.Compare):
            left = _convert(node.left)
            ops = [op_map[type(op)] for op in node.ops]
            comparators = [_convert(comp) for comp in node.comparators]
            return ops[0](left, comparators[0])
        elif isinstance(node, ast.BoolOp):
            values = [_convert(v) for v in node.values]
            return op_map[type(node.op)](*values)
        elif isinstance(node, ast.Name):
            return pl.col(node.id)
        elif isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Str):
            return node.s
        else:
            raise ValueError(f"不支持的表达式节点: {type(node)}")

    try:
        tree = ast.parse(filter_rule, mode='eval')
        return _convert(tree.body)
    except Exception as e:
        raise ValueError(f"无法解析过滤规则: {e}")


def filter_algo(params: list, props: list, start_date='', end_date=''):
    df = params[0]
    filter_rule = params[1]
    if isinstance(df, pl.DataFrame) and df.is_empty():
        return pl.DataFrame()
    try:
        # 使用安全的表达式转换
        expr = safe_eval_filter_rule(filter_rule)
        f_df = df.filter(expr)
    except Exception as ex:
        logger.info(f"filter数据出错: {ex}")
        try:
            all_filter_fields = rule_engine.extract_fields(filter_rule)
            rule_expr = rule_engine.build_expression(filter_rule, all_filter_fields)
            f_df = df.filter(rule_expr[0])
        except Exception as exp:
            raise exp
    return f_df

if __name__ == '__main__':
    print(asset_tag([['债券', '股票']], ['修正久期']))
    print(issuer_tag([['主体评级AAA']], ['流通市值']))
    print(port_tag([['现金管理类']], ['万份收益', '单位净值', '净资产']))




import muse.muse_lang as ml
import muse.sample as sample
# from muse.data_source.test_ds import TestDataSource

def test_code(code):
    # ml.init_ds(TestDataSource())
    print(ml.ast_check(code))
    return ml.trial(code)

if __name__ == '__main__':
    result = test_code(sample.demo1)
    if result['code'] == 'SUCCESS':
        print(result['data'])
        print(result['error_msgs'])
    else:
        print(result['error_msgs'])